---- Mod Mangement
-- Only for internal version
--
-- Always check for presence before use.
--
-- @author Oskar Eisemuth
-- @copyright 2018-2021
-- @module commonapi._modmm

local modmm = {
}

require "serialize"

local _utils = commonapi.utils
local _log = _utils.logPrep("commonapi._modmm")
local _native = commonapi._native

function modmm.getDirFor(name)
	local userpath = _native.paths.getUserPath()
	if (userpath == nil) then
		_log("No UserPath found?")
		return
	end
	userpath = string.gsub(userpath, '\\', '/')
	if (userpath:sub(-1) == '/') then
		userpath = userpath:sub(1, -2)
	end
	local path = userpath .. '/commonapi/'..name
	_native.fs.mkdir(path)
	return path
end


function modmm.getDirLocalMods()
	local userpath = _native.paths.getUserPath()
	if (userpath == nil) then
		_log("No UserPath found?")
		return
	end
	userpath = string.gsub(userpath, '\\', '/')
	if (userpath:sub(-1) == '/') then
		userpath = userpath:sub(1, -2)
	end
	local path = userpath .. '/mods'
	_native.fs.mkdir(path)
	return path
end

function modmm.getDirGameMods()
	local userpath = _native.paths.getGamePath()
	if (userpath == nil) then
		_log("No getGamePath found?")
		return
	end
	userpath = string.gsub(userpath, '\\', '/')
	if (userpath:sub(-1) == '/') then
		userpath = userpath:sub(1, -2)
	end
	local path = userpath .. '/mods'
	_native.fs.mkdir(path)
	return path
end


function modmm.runprocess(pcmd)
	_log("Trying to run "..pcmd)
	if (commonapi._info.sys == 1) then
		if (_native.utils.runprocess ~= nil) then
			local ok, msg = _native.utils.runprocess(pcmd)
			return ok, msg
		end
		pcmd = "\"pcmd\""
	end

	local p = assert(io.popen(pcmd, "r"))
	if (p ~= nil) then
		local msg = p:read("*a")
		p:close()
		return true, msg
	end
	return false, "Can't run "..pcmd
end


local function splitnewlines(str)
	if string.find(str, "\n") == nil then
		return { str }
	end

	local lines = {}
	local pat = "(.-)\r?\n()"
	local lastPos
	for part, pos in string.gmatch(str, pat) do
		lines[#lines+1] = part
		lastPos = pos
	end
	lines[#lines+1] = string.sub(str, lastPos)
	return lines
end

function modmm.runprocessGetLines(pcmd)
	_log("Trying to run "..pcmd)
	if (commonapi._info.sys == 1) then
		if (_native.utils.runprocess ~= nil) then
			local ok, msg = _native.utils.runprocess(pcmd)
			if (ok == true) then
				return true, splitnewlines(msg)
			end
			return ok, msg
		end
		pcmd = "\"pcmd\""
	end
	local lines = {}
	local p = io.popen(pcmd)
	if (p ~= nil) then
		for line in p:lines() do
			lines[#lines+1] = line
		end
		return true, lines
	end
	return false, "Can't run "..pcmd
end

function modmm.download(filename, url, download_size)
	local pcmd
	local q_file = _utils.quoteParam(filename)
	local q_url = _utils.quoteParam(url)


	if (commonapi._info.sys == 1) then
		pcmd = 'curl -f  -w "HTTP Status: %{http_code}" -sS -m 30 --create-dirs -z '..q_file..' -o '..q_file..' '..q_url
	else
		--pcmd = '/usr/bin/curl --create-dirs -z '..q_repolocalfile..' -o '..q_repolocalfile..' '..q_repositoryurl
		pcmd = 'LD_LIBRARY_PATH=; LD_PRELOAD=; /usr/bin/curl -f  -w "HTTP Status: %{http_code}" -sS -m 30 --create-dirs -z '..q_file..' -o '..q_file..' '..q_url ..' 2>&1'
	end

	local ok, msg = modmm.runprocess(pcmd)
	if (ok == false) then
		return false, msg
	end

	local statuscode = msg.match(msg, "HTTP Status: (%d+)")
	if (statuscode ~= nil and (statuscode == "200" or statuscode == "304")) then
		if (download_size ~= nil) then
			local stat = _native.fs.stat(filename);
			if (download_size ~= stat.size) then
				 msg = _("Download incomplete: ${filename}\nshould be ${expectedsize} but file is ${currentsize}") % {
					filename = filename,
					expectedsize = download_size,
					currentsize = stat.size,
				};
				return false, msg
			end
		end

		return true
	end


	return false, msg
end


function modmm.downloadRepository(filename, repositoryurl)
	--local repositoryurl = "https://www.transportfever.net/filebase/repos/tpf2.json"
	local path = modmm.getDirFor("repos")
	if (path == nil) then
		return false
	end
	local repolocalfile = path.."/"..filename

	local ok, errormsg = modmm.download(repolocalfile, repositoryurl)

	if (ok == false) then
		return false, errormsg
	end

	ok, errormsg = modmm.convertRepositoryFilename(filename)
	if (ok == false) then
		return false, errormsg
	end

	return true
end




local function compareversion(a, b)
	local l = math.max(#a.version, #b.version)
	for i = 1, l do
		local ai = a.version[i] or 0
		local bi = b.version[i] or 0
		if (ai < bi) then
			return false
		end
		if (ai > bi) then
			return true
		end
	end
	return false
end

function modmm._repoConverterToModIdMap(files)
	local mapidtoversion = {}
	for m, fileentry in ipairs(files) do
		if (fileentry.modid ~= nil) then
			local feversion = _utils.convertVersion(fileentry.version)
			fileentry.version = feversion

			if (mapidtoversion[fileentry.modid] == nil) then
				mapidtoversion[fileentry.modid] = {}
			end
			table.insert(mapidtoversion[fileentry.modid], fileentry)
		end
	end

	for m, entry in pairs(mapidtoversion) do
		table.sort(entry, compareversion)
	end
	return mapidtoversion
end


function modmm.convertRepositoryFilename(filename)
	local dkjson = require("commonapi2.extlib.dkjson")
	local jsonstr

	local path = modmm.getDirFor("repos")
	if (path == nil) then
		return
	end
	local fullfilename = path..'/'..filename
	local jsonfile = io.open(fullfilename, "r")
	if (jsonfile == nil) then
		return false,  "Can't open"..fullfilename
	end

	jsonstr = jsonfile:read("*a")
	jsonfile:close()


	local repo, pos, err = dkjson.decode(jsonstr, 1, nil, nil)
	if err then
		print ("Error:", err)
		return false, err
	end

	if (type(repo.files) == "table") then
		local mapidtoversion = modmm._repoConverterToModIdMap(repo.files)

		local data = {
			repo = repo.repo,
			mods = mapidtoversion,
		}

		local marshalfile = io.open(fullfilename..'.m', "wb")
		if (marshalfile == nil) then
			return false, "Can't open"..fullfilename..'.m'
		end
		marshalfile:write(_native.marshal.encode(data))
		marshalfile:close()
	end

	return true
end

function modmm.getRepositoryData(filename)
	local path = modmm.getDirFor("repos")
	if (path == nil) then
		return
	end
	local fullfilename = path..'/'..filename

	local marshalfile = io.open(fullfilename..'.m', "rb")
	if (marshalfile == nil) then
		return
	end
	local marshaldata = marshalfile:read('*a')
	marshalfile:close()
	local repo = _native.marshal.decode(marshaldata)


	local repomods = repo.mods or {}
	local repodata = repo.repo or {}

--	local starttime = os.clock()
	for modid, entry in pairs(repomods) do
		for idx, repomodentry in ipairs(entry) do
			repomodentry.repodata = repodata
		end
	end
--	print(string.format("elapsed time: %.2f\n", os.clock() - starttime))

	return repo
end


function modmm.applyRepositoryFilename(filename, mods)
	--local starttime = os.clock()

	local repo = modmm.getRepositoryData(filename)
	if (repo == nil) then
		return
	end

	local repomods = repo.mods or {}
	local repodata = repo.repo or {}

	for m, entry in ipairs(mods) do
		if (repomods[entry.id] ~= nil) then
			local repomodentries = repomods[entry.id]
			--for x, repomodentry in ipairs(repomodentries) do
			--	repomodentry.repodata = repodata
			--end
			local upgradeversion = repomodentries[1].version
			entry.repomods = repomodentries
			--entry.repodata = repodata
			if (_utils.compareVersion(entry.version, upgradeversion) < 0) then
				entry.upgradeversion = upgradeversion
			end
		end
	end

	--print(string.format("elapsed time: %.2f\n", os.clock() - starttime))
	return repodata.utc_time or nil
end



function modmm.selectInstallDir(fullmodid)
	local modsdir = modmm.getDirLocalMods()
	if (modsdir == nil) then
		return false, "Can't find local mod dir"
	end

	local modsgamedir = modmm.getDirGameMods()
	if (modsgamedir == nil) then
		return false, "Can't find game mod dir"
	end

	local destdir = modsdir..'/'..fullmodid
	local destdirgame = modsgamedir..'/'..fullmodid

	local destdir_real = _native.fs.realpath(destdir)
	if (destdir_real ~= nil) then
		_log("Found old mod @ "..destdir_real)
	else
		local destdirgame_real = _native.fs.realpath(destdirgame)
		if (destdirgame_real ~= nil) then
			_log("Found old mod @ "..destdirgame_real)
			destdir = destdirgame
		end
	end
	return true, destdir
end

function modmm.downloadAndInstallMod(repoentry)
	local ok, errormsg, destdir
	local downloaddir = modmm.getDirFor("downloads")
	if (downloaddir == nil) then
		return false, "Can't create downloads"
	end

	local modstoragedir = modmm.getDirFor("modstorage")
	if (modstoragedir == nil) then
		return false, "Can't create modstorage"
	end

	if (repoentry.modid == nil) then
		return false, "Internal error in repository, modid not found"
	end

	ok, destdir = modmm.selectInstallDir(repoentry.modid)
	if (ok == false) then
		return false, destdir
	end

	local downloadurl = repoentry.repodata.download_prefix_url..repoentry.download
	local strversion = 	''..table.concat(repoentry.version or '1', '.')
	local downloadextract =  downloaddir..'/'..repoentry.modid..'__'..strversion
	local downloadfile = downloadextract..'.package'

	local download_size = repoentry.download_size or nil
	ok, errormsg = modmm.download(downloadfile, downloadurl, download_size)

	if (ok == false) then
		return false, errormsg
	end

	ok, errormsg = modmm._doInstallMod(repoentry.modid, downloadfile, downloadextract, modstoragedir, destdir)

	if (ok == false) then
		return false, errormsg
	end
end


function modmm._doInstallMod(fullmodid, downloadfile, downloadextract, modstoragedir, destdir)
	local ok, errormsg = modmm.extractFile(downloadfile, fullmodid, downloadextract)
	if (ok == false) then
		return false, errormsg
	end
	local extractmoddir = _utils.fixslashes(downloadextract..'/'..fullmodid)

	if (_native.fs.realpath(extractmoddir) == nil) then
		return false, _("7-zip extract ${src} to ${dest} failed\n") % {
			src = downloadfile,
			dest = extractmoddir
		}
	end

	local destdirfixed = _utils.fixslashes(destdir)
	local destdir_real = _native.fs.realpath(destdirfixed)
	if (destdir_real ~= nil) then
		local savedfolder = _utils.fixslashes(modstoragedir..'/'..fullmodid..'-saved-'..os.time())
		_log("Moving old mod folder to new dest:\n"..destdir_real.."\n"..savedfolder);
		errormsg = _native.fs.rename(destdir_real, savedfolder)
		if (errormsg ~= nil) then
			return false, "Moving old mod folder failed:\n"..errormsg
		end
	end

	_log("Moving extracted folder to new dest:\n"..extractmoddir.."\n"..destdirfixed);
	errormsg = _native.fs.rename(extractmoddir, destdirfixed)
	if (errormsg ~= nil) then
		return false, "Moving extracted folder to new dest failed:\n"..errormsg.."\n"..extractmoddir.." -> \n"..destdirfixed
	end
	return true
end


function modmm.installModFromArchiv(filename)
	local ok, errormsg, fullmodid, destdir

	local downloaddir = modmm.getDirFor("downloads")
	if (downloaddir == nil) then
		return false, "Can't create downloads"
	end

	local modstoragedir = modmm.getDirFor("modstorage")
	if (modstoragedir == nil) then
		return false, "Can't create modstorage"
	end

	ok, fullmodid = modmm.getFileInfo(filename)
	if (ok == false) then
		return false, fullmodid
	end

	ok, destdir = modmm.selectInstallDir(fullmodid)
	if (ok == false) then
		return false, destdir
	end

	local downloadextract =  downloaddir..'/local-'..fullmodid..'__'..os.time()

	ok, errormsg = modmm._doInstallMod(fullmodid, filename, downloadextract, modstoragedir, destdir)
	if (ok == false) then
		return false, errormsg
	end
	return true
end


function modmm.extractFile(filename, direntry, extractto)
	local q_file = _utils.quoteParam( _utils.fixslashes(filename) )
	local q_direntry = _utils.quoteParam( _utils.fixslashes(direntry) )
	local q_extractto = _utils.quoteParam( _utils.fixslashes(extractto) )

	local q_cmd = 'x '..q_file..' -o'..q_extractto..' '..q_direntry..' -r -aoa -bd -bso0 -bsp0 -bse1'

	local pok, pcmd = modmm.getExtractCommand(q_cmd)
	if (pok ~= true) then
		return false, pcmd
	end


	local ok, msg = modmm.runprocess(pcmd)
	if (ok == false) then
		return false, msg
	end
	print(" modmm.extractFile>>>")
	print(msg)
	print("<<<modmm.extractFile")
	if (msg ~= "") then
		return false, msg
	end

	local extractmoddir = _utils.fixslashes(extractto..'/'..direntry)

	if (_native.fs.realpath(extractmoddir) == nil) then
		return false, "Extract failed"
	end
--[[
	local p = io.popen(pcmd)
	if (p ~= nil) then
		local msg = p:read("*a")
		print(msg)
		p:close()
		if (msg ~= "") then
			return false, msg
		end
	end
--]]
	return true
end



function modmm.getExtractorBin()
	local bin7zip
	if (commonapi._info.sys == 1) then
		bin7zip = "C:\\Program Files\\7-Zip\\7z.exe"

		local dir7zip = modmm.getDirFor("7zip")
		local bin7zipDownload = dir7zip.."/Files/7-Zip/7z.exe"

		local bin7zip_real = _native.fs.realpath(bin7zipDownload)
		if (bin7zip_real ~= nil) then
			bin7zip = bin7zip_real
		end
	else
		bin7zip = "/usr/bin/7z"
	end

	local usersevenzipbin = commonapi._info.settings.location7zip

	if (type(usersevenzipbin) == 'string' and usersevenzipbin ~= "") then
		bin7zip = usersevenzipbin
	end

	_log("Using 7zip @ ", bin7zip)
	return bin7zip
end

function modmm.extractorIsUsable()
	local bin7zip =	modmm.getExtractorBin()

	local bin7zip_real = _native.fs.realpath(bin7zip)
	if (type(bin7zip_real) ~= 'string') then
		return false, "Not found"
	end
	local pcmd = bin7zip_real -- _utils.quoteParam(bin7zip_real)

	local ok, plinesormsg = modmm.runprocess(pcmd)
	if (ok == false) then
		return false, "Can't start"
	end

	if (plinesormsg ~= nil) then
		local version = plinesormsg:match("7%-Zip%s+([0-9.]+)")
		if (version ~= nil) then

			local xversion = _utils.convertVersion(version)

			if (_utils.compareVersion(xversion, {19, 0}) >= 0) then
				_log("7-Zip version", version)
				return true, version
			end
			_log("7-Zip version to old", version)
			return false, "7-Zip version to old"
		end
	end

	return false, "Can't get 7-zip version"
end


function modmm.extractorDownload()
	local msi7zip = "7z1900-x64.msi"
	local msi7zipdl = "https://7-zip.org/a/"..msi7zip
	local dir7zip = modmm.getDirFor("7zip")
	if (dir7zip == nil) then
		return false, "Can't create 7zip dir"
	end
	local downloaddir = modmm.getDirFor("downloads")
	if (downloaddir == nil) then
		return false, "Can't create downloads"
	end
	local ok, errormsg = modmm.download(downloaddir..'/'..msi7zip, msi7zipdl)

	if (ok == false) then
		return false, errormsg
	end

	local q_file = _utils.quoteParam(  _utils.fixslashes( downloaddir..'/'..msi7zip ) )
	local q_direntry = _utils.quoteParam(  _utils.fixslashes( dir7zip..'/' ) )
	local pcmd = "msiexec /a "..q_file.." TARGETDIR=".. q_direntry .. " /qn"

	local ok, plinesormsg = modmm.runprocessGetLines(pcmd)
	if (ok == false) then
		return false, plinesormsg
	end
	return true
end

function modmm.getExtractCommand(q_cmd)
	local pcmd
	local bin7zip = modmm.getExtractorBin()

	local bin7zip_real = _native.fs.realpath(bin7zip)
	if (type(bin7zip_real) ~= 'string') then
		local err = _("7zip not found in ${filepath}") % {
			filepath = bin7zip
		}
		return false, err
	end

	local q_7zip =  _utils.quoteParam(bin7zip_real)
	if (commonapi._info.sys == 1) then
		--pcmd = '"'..q_7zip..' '..q_cmd..'"'
		pcmd = q_7zip..' '..q_cmd
	else
		pcmd = 'LD_LIBRARY_PATH=; LD_PRELOAD=; '..q_7zip..' '..q_cmd
	end
	return true, pcmd
end



function modmm.getFileInfo(filename)
	local REGEX_MODFOLDER = "^([a-zA-Z0-9_-]+_([0-9]+))%/mod%.lua$"
	local q_file = _utils.quoteParam( _utils.fixslashes(filename) )
	local q_direntry = _utils.quoteParam('*/mod.lua')


	local q_cmd = ' l -slt -bd -bso0 -bsp0 -bse1 '..q_file..' '..q_direntry
	
	local pok, pcmd = modmm.getExtractCommand(q_cmd)
	if (pok ~= true) then
		return false, pcmd
	end

--	_log("Trying to run "..pcmd)

	local modid, modmajorversion
	local lines = {}

	local ok, plinesormsg = modmm.runprocessGetLines(pcmd)

	if (ok == false) then
		return ok, plinesormsg
	end
	local msg = table.concat(plinesormsg, "\n")
	local files = modmm.parse7zipoutput(plinesormsg)

	for i, file in ipairs(files) do
		modid, modmajorversion = file:match(REGEX_MODFOLDER)
		if (modid ~= nil) then
			print("Found modid"..modid)
			break;
		end
	end
	if (modid ~= nil) then
		return true, modid
	end

	if (msg ~= "") then
		return false, msg
	end
--[[
	local p = io.popen(pcmd)
	if (p ~= nil) then
		for line in p:lines() do
			lines[#lines+1] = line
		end

		p:close()

		msg = table.concat(lines, "\n")
		--commonapi.dmp(lines)
		--print(msg)
		local files = modmm.parse7zipoutput(lines)

		for i, file in ipairs(files) do
			modid, modmajorversion = file:match(REGEX_MODFOLDER)
			if (modid ~= nil) then
				print("Found modid"..modid)
				break;
			end
		end
		if (modid ~= nil) then
			return true, modid
		end

		if (msg ~= "") then
			return false, msg
		end
	end
--]]

	return false, "Can't get modid from "..filename
end


function modmm.parse7zipoutput(lines)
	local files = {}
	local startoflist = '----------'
	local pathline = 'Path = '
	local inheader = true
	for i, line in ipairs(lines) do
		if (inheader) then
			if (line == startoflist) then
				inheader = false
			end
		else
			if (line:sub(1, #pathline) == pathline) then
				files[#files+1] = string.gsub( line:sub(#pathline+1), '\\', '/')
			end
		end
	end
	commonapi.dmp(files)

	return files
end




local modmmrepos = {
	repolistDefault = {
		{ filename = "tpfnet-tpf2.json",  url = "https://www.transportfever.net/filebase/repos/tpf2.json" }
	},
	repolist = {
		{ filename = "tpfnet-tpf2.json",  url = "https://www.transportfever.net/filebase/repos/tpf2.json" }
	},
	loadedrepos = {},
	utc_time = nil,
}


function modmmrepos:reload()
	self:_loadRepoList()

	self.utc_time = nil
	self.loadedrepos = {}
	local utc_time = 0
	for idx, repolistentry in ipairs(self.repolist) do
		local repo = modmm.getRepositoryData(repolistentry.filename)
		if (repo ~= nil) then
			self.loadedrepos[#self.loadedrepos+1] = repo
			local repodata = repo.repo or {}
			if (repodata.utc_time ~= nil) then
				repolistentry.lastupdate = repodata.utc_time
				if (repodata.utc_time > utc_time) then
					utc_time = repodata.utc_time
				end
			end
		else
			_log("Can't load "..repolistentry.filename)
		end
	end

	self.utc_time = utc_time or nil
end

function modmmrepos:_loadRepoList()
	local path = modmm.getDirFor("repos")
	if (path == nil) then
		return false
	end
	local repolistlocalfile = path.."/repolist.lua"

	local resultfn, err = loadfile(repolistlocalfile, "bt", {})
	if (resultfn ~= nil) then
		local data = resultfn()
		if (type(data) == "table") then
			self.repolist = data
			return
		end
	end

	self.repolist = _utils.deepcopy(self.repolistDefault)
end

function modmmrepos:setRepoList(newList)
	local path = modmm.getDirFor("repos")
	if (path == nil) then
		return false
	end
	local repolistlocalfile = path.."/repolist.lua"
	local fh = io.open(repolistlocalfile, "w")
	if (fh) then
		local data = _utils.serializeData(newList)
		fh:write(data)
		fh:close()
	end
	self:_loadRepoList()
end
function modmmrepos:addRepoToList(repopfilename, repourl)
	self.repolist[#self.repolist+1] = {
		filename = repopfilename, url = repourl
	}
	self:setRepoList(self.repolist)
end


function modmmrepos:removeRepoFromList(repofilename)
	local newlist = {}
	for idx, repolistentry in ipairs(self.repolist) do
		if (repolistentry.filename ~= repofilename) then
			newlist[#newlist+1] = repolistentry
		end
	end
	self:setRepoList(newlist)
end


function modmmrepos:getRepoList()
	return self.repolist
end

function modmmrepos:getLastRepoDataUpdate()
	return self.utc_time or nil
end


function modmmrepos:updateRepoData()
	local ok, errormsg
	for idx, repolistentry in ipairs(self.repolist) do
		ok, errormsg = modmm.downloadRepository(repolistentry.filename, repolistentry.url)
		if (ok == false) then
			return ok, errormsg
		end
	end
	self:reload()
	return true
end


function modmmrepos:findEntriesByModId(modid)
	local ret = nil
	for idx, loadedrepo in ipairs(self.loadedrepos) do
		local repomods = loadedrepo.mods or {}
		if (repomods[modid] ~= nil) then
			ret = repomods[modid]
		end
	end
	return ret
end


function modmmrepos:applyRepositoriesToMods(mods)
	--local starttime = os.clock()

	for idx, loadedrepo in ipairs(self.loadedrepos) do
		local repomods = loadedrepo.mods or {}

		for m, entry in ipairs(mods) do
			if (repomods[entry.id] ~= nil) then
				local repomodentries = repomods[entry.id]
				local upgradeversion = repomodentries[1].version
				entry.repomods = repomodentries
				--entry.repodata = repodata
				if (_utils.compareVersion(entry.version, upgradeversion) < 0) then
					entry.upgradeversion = upgradeversion
				end
			end
		end
	end
	--print(string.format("elapsed time: %.2f\n", os.clock() - starttime))

end


function modmm.getRepoDataObj()
	return modmmrepos
end



return modmm