local LineExtraDataWindow = {}

local _log = commonapi.utils.logPrep("commonapi2.ui.LineExtraDataWindow")
local _utils = commonapi.utils
local _ = _utils._t
local _syncdata = commonapi._syncdata
local ui = commonapi.ui

--commonapi._reloadModule("_modmm")

LineExtraDataWindow = ui.Window:new()

LineExtraDataWindow.id = "LineExtraDataWindow"
LineExtraDataWindow.height = 500
LineExtraDataWindow.width = 600

LineExtraDataWindow.caption = _("Line Destination")
LineExtraDataWindow.data = {}

local linesextradetails_sync =
	_syncdata.SyncObject:new(
	{
		key = "linesextradetails",
		data = {}
	}
)

function LineExtraDataWindow.create()
	local o
	o = LineExtraDataWindow:new()
	o:refresh()
	o:open()
	return o
end

function LineExtraDataWindow:onchange()
end

function LineExtraDataWindow:saveData()
	_log("Doing datasync")
	linesextradetails_sync:syncData()
	commonapi.dmp(linesextradetails_sync:getData())
	if (commonapi._native.LabelSystem ~= nil) then
		commonapi._native.LabelSystem.updateData(linesextradetails_sync:getData())
	end
	local datasyncresult = runInGUIThread("return commonapi_gamescript.saveData()")
end

function LineExtraDataWindow:refresh()
	local uilist = {}
	--	print("LineExtraDataWindow:refresh")
	--	commonapi.dmp(self.params)
	--	print("/LineExtraDataWindow:refresh")
	local linedata = runInGUIThread("return commonapi_gamescript.getLineData()")

	uilist[#uilist + 1] = {
		typ = "SmallButton",
		id = "checkUpdates",
		caption = _("Refresh"),
		sameline = true,
		onclick = function(item, win)
			win:refresh()
		end
	}
	local uiScrollList = {}
	uilist[#uilist + 1] = {
		typ = "ScrollRegion",
		id = "ScrollRegion",
		border = false,
		size = {
			0,
			-200
		},
		uilist = uiScrollList
	}

	for m, line in pairs(linedata) do
		uiScrollList[#uiScrollList + 1] = {
			--typ = "Label",
			typ = "_Selectable",
			_fix_overlap = true,
			id = line.id .. "__item",
			caption = line.name .. "  (" .. line.id .. ")",
			--column = 0,
			--columnwidth = 200.0,
			--rownew = true,
			data = {
				lineId = line.id
			},
			onclick = function(item, win)
				win.data["curentry"] = item.data.lineId
				win:genLineDetails(item.data.lineId)
			end,
			selected = function(item, win)
				return (win.data["curentry"] == item.data.lineId)
			end
		}

		uiScrollList[#uiScrollList + 1] = {
			typ = "SmallButton",
			_fix_overlap = true,
			id = line.id .. "__item_reset",
			caption = "Reset",
			sameline = true,
			data = {
				lineId = line.id
			},
			onclick = function(item, win)
				win.data["curentry"] = nil
				if (win.uilist[#win.uilist].typ == "TabBar") then
					win.uilist[#win.uilist] = nil
				end
				local linesextradetails = linesextradetails_sync:getDataOrInit({})
				if (linesextradetails.lines ~= nil) then
					linesextradetails.lines[item.data.lineId] = nil
				end
				win:saveData()
			end,
		}
	end

	self.uilist = uilist
end




function LineExtraDataWindow:genLineDetails(lineid)
	if (self.uilist[#self.uilist].typ == "TabBar") then
		self.uilist[#self.uilist] = nil
	end
	self.params = {}
	--commonapi.dmp(lineid)
	local linedata = runInGUIThread("return commonapi_gamescript.getLineData()")

	local linesextradetails = linesextradetails_sync:getDataOrInit({})

	--local line = linedata[lineid] or nil
	local line
	for m, curline in pairs(linedata) do
		if (curline.id == lineid) then
			line = curline
		end
	end

	if (line == nil) then
		_log("linedata is nil?", lineid)
		return
	end

	if (linesextradetails.version ~= 1) then
		linesextradetails.lines = {}
		linesextradetails.version = 1
	end

	if (linesextradetails.lines == nil) then
		linesextradetails.lines = {}
	end


	if (linesextradetails.lines[lineid] == nil) then
		linesextradetails.lines[lineid] = {}
	end

	local linedetails = linesextradetails.lines[lineid]

	if (linedetails.endStations == nil) then
		linedetails.endStations = {}
	end

	local uiTabs = {}

	local uiTabBar = {
		typ = "TabBar",
		id = "LineDetailsTab",
		uilist = uiTabs
	}

	local uilistInfoStops = {}

	local columnsStops = {
		{"Stop", 0},
		{"EntityId", self.width - 100 - 80},
		{"Endstop", self.width - 100}
	}
	local columnsStopsObj = {
		typ = "Columns",
		count = #columnsStops,
		border = true
	}
	uilistInfoStops[#uilistInfoStops + 1] = columnsStopsObj

	for i, column in ipairs(columnsStops) do
		uilistInfoStops[#uilistInfoStops + 1] = {
			typ = "Label",
			column = i - 1,
			caption = column[1],
			columnoffset = column[2]
		}
	end


	local endStationsVisited = {}
	local endStations = linedetails.endStations

	for stopIdx = 1, #line.stops_data do
		local stop = line.stops_data[stopIdx]

		if (endStationsVisited[stop.id] == nil) then
			endStationsVisited[stop.id] = 1
		else
			endStationsVisited[stop.id] = endStationsVisited[stop.id] + 1
		end

		local uniqueId = stop.id..'-'..endStationsVisited[stop.id];

		local idprefix = line.id .. "." .. stopIdx 

		uilistInfoStops[#uilistInfoStops + 1] = {
			typ = "Label",
			column = 0,
			id = idprefix .. "__label",
			caption = stop.name
		}
		uilistInfoStops[#uilistInfoStops + 1] = {
			typ = "Label",
			column = 1,
			id = idprefix .. "__lentityid",
			caption = stop.id
		}

		local idCheckbox = idprefix .. "__isEndstop"
		local idText = idprefix .. "__EndStopCaption"
		uilistInfoStops[#uilistInfoStops + 1] = {
			typ = "Checkbox",
			column = 2,
			id = idCheckbox,
			caption = "",
			data = {
				uniqueId = uniqueId
			},
			onclick = function(item, win)
				if (win.params[item.id] == true) then
					endStations[item.data.uniqueId] = {
						caption = win.params[idText] or "",
					}
				else
					endStations[item.data.uniqueId] = nil
				end
				win:saveData()
			end
		}

		uilistInfoStops[#uilistInfoStops + 1] = {
			typ = "TextInput",
			column = 2,
			sameline = true,
			id = idText,
			caption = "",
			data = {
				uniqueId = uniqueId
			},
			onchange = function(item, win)
				win.params[idCheckbox] = true

				endStations[item.data.uniqueId] = {
					caption = win.params[idText] or "",
				}
				win:saveData()
			end
		}

		self.params[idCheckbox] = false
		self.params[idText] = ""

		if (endStations[uniqueId] ~= nil) then
			self.params[idCheckbox] = true
			if (endStations[uniqueId].caption ~= nil) then
				self.params[idText] = endStations[uniqueId].caption
			end
		end
	end


	uiTabs[#uiTabs + 1] = {
		typ = "TabItem",
		id = "LineDetailsStops",
		caption = "Stops",
		uilist = {
			{
				typ = "ScrollRegion",
				id = "ScrollRegion",
				size = {
					0,
					0
				},
				uilist = uilistInfoStops
			}
		}
	}

	self.uilist[#self.uilist + 1] = uiTabBar
end

local function menucallback()
	LineExtraDataWindow.create()
end

local function reloadcallback()
	package.loaded["commonapi2.ui.LineExtraDataWindow"] = nil
	print("Reload LineExtraDataWindow_open callback called")
	local LineExtraDataWindowObj = require "commonapi2.ui.LineExtraDataWindow"
	LineExtraDataWindowObj.run()
end

function LineExtraDataWindow.run()
	_log("LineExtraDataWindow.run")
	ui.createModMenuEntry("LineExtraDataWindow_open", "Line Destination", menucallback)
	ui.createModMenuEntry("LineExtraDataWindow_reload", "Reload Line Destination", reloadcallback)
end

return LineExtraDataWindow
